/***************************************************************************
© F. Rousset, R. Leblois, C.R. Beeravolu 2012-

francois.rousset@univ-montp2.fr
raphael.leblois@mnhn.fr

This file is part of Migraine. This software is a computer program
whose purpose is to perform statistical analyses.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author, the holder of the
economic rights, and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading, using, modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate, and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and, more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

 ***************************************************************************/

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <cstring>
#include <cstdlib>
#include <vector>
#include <map>
#include <numeric>

#include "genepop.h"
#include "migraine.h"
#include "nexus.h"
#include "hypercube.h"

CNexus::CNexus()
{
	fileName = globalNexusFilename;
	seqNum = seqSize = hapCount = currInfSites = origInfSites = (size_t) 0;
    hyperSegCount = samSize = locNum = 0;
	nexusFileOK = false;
	recodeISMBool = false;
	pass4GTBool = true;
}

void CNexus::cleanNEXUSFile(const int &locusNum)
{
/*
 	 check for the existence of the NEXUS file, its EOF/EOL types
 */

	locNum = locusNum;
	{
		stringstream inStst, outStst;
		if (totalLociCount > 1) {
			inStst << "loc" << locNum << ".nex";
			outStst << "output_seqloc_job" << _mls << "_" << locNum << ".txt";
		}
		else {
			inStst << ".nex";
			outStst << "output_seqloc_job" << _mls << ".txt";
		}
		fileName += inStst.str();
		outFilename = outStst.str();
	}

	//	Checks for the End-of-Line characters
	int rc = set_eof_check_EOLtype(fileName, false);
	//rc indicates file not found
	if (rc == -1) {
		cerr<<"... now exiting";
		if (cinGetOnError)
			cin.get();
		exit(-1);
	}

	NexusMsgStr << "\nNEXUS file \"" << fileName << "\"....found" << endl;

	//	Attempts to convert the input file EOL format to that of the UNIX environment
	set_UNIX_EOLtype(fileName);

	nexusFileOK = true;
}

void CNexus::readNEXUSFile(CSample *gpSample)
{
/*
	Reads ONLY the data block of a NEXUS file and performs a few checks on the extracted data
*/

	string str1, str2;
	string dataOptions, seqDatatype, seqSymbols, seqGapSym, seqMisSym;
	size_t pos, beginBlock, endBlock;
	ifstream nexusFile;
	vector<string> fileBuf;

	//	a safeguard for ensuring the execution order of CNexus procedures during the NEXUS file read
	if (!nexusFileOK) {
		cerr << "\nThe Nexus file \"" << fileName << "\" needs to be cleaned by the procedure CNexus::cleanNEXUSFile before being read!"
			 << "\nAborting Migraine...Press any key to exit.\n";
		if (cinGetOnError)
			cin.get();
		exit(-1);
	}

	//	start file read
	nexusFile.open(fileName.c_str(), std::ios::in);

	//	Extracts and checks for "#NEXUS" on the first line, if not found then abort the program
	lineBuf.clear();
	getline(nexusFile, lineBuf);
	{
		stringstream tokenSStr;
		tokenSStr.str(lineBuf);
		tokenSStr >> lineBuf;
	}
	if (cmp_nocase(lineBuf, "#NEXUS") != 0) {
		cerr << "\nError! Migraine couldn't recognize the header \"" << lineBuf << "\"\n"
			 << "from The input file \"" << fileName << "\". It may not be of type NEXUS.\n";
		if (cinGetOnError) {
			cerr << "\nIgnoring...Press any key to continue...\n";
			cin.get();
			//exit(-1);
		}
	}

	NexusMsgStr << "NEXUS file header....OK" << endl;

	//	reads the input file, character by character, into a single string
	lineBuf.clear();
	while (!nexusFile.eof())
		lineBuf += nexusFile.get();

	//	end file read
	nexusFile.close();

	//	removes all the comments from the "single-lined file"
	removeNEXUSComments();

	NexusMsgStr << "NEXUS comments(if any)....removed" << endl;

	//	splits the uncommented NEXUS file using the ";" token and stores it in a string vector
	{
		stringstream tokenSStr;
		tokenSStr.str(lineBuf);

		lineBuf.clear();
		getline(tokenSStr, lineBuf, ';');
		while (!tokenSStr.eof()) {
			fileBuf.push_back(lineBuf);
			lineBuf.clear();
			getline(tokenSStr, lineBuf, ';');
		}
	}

	//	Indexes the start and end positions of the NEXUS data block in the string vector
	beginBlock = endBlock = 0;
	while (beginBlock <= fileBuf.size() - 1) {
		stringstream tokenSStr1;
		tokenSStr1.str(fileBuf[beginBlock]);
		tokenSStr1 >> str1 >> str2;
		if ((cmp_nocase(str1,"begin") == 0) && ((cmp_nocase(str2,"data") == 0) || (cmp_nocase(str2,"characters") == 0))) {
			string str3;
			endBlock = beginBlock;
			while (endBlock <= fileBuf.size() - 1) {
				stringstream tokenSStr2;
				tokenSStr2.str(fileBuf[endBlock]);
				tokenSStr2 >> str3;
				if (cmp_nocase(str3,"end") == 0)
					break;
				++endBlock;
			}
			break;
		}
		++beginBlock;
	}
	//	IF the data block delimiters have not been found then quit program
	if (beginBlock == fileBuf.size()) {
		cerr << "\nError! No data found.\n"
			 << "Migraine couldn't find the data block in \"" << fileName << "\"."
			 << "\nAborting Migraine...Press any key to exit.\n" << endl;
		if (cinGetOnError)
			cin.get();
		exit(-1);
	}
	else if (endBlock == fileBuf.size()) {
		cerr << "\nError! The NEXUS file data block ended abnormally in \"" << fileName << "\"."
			 << "\nAborting Migraine...Press any key to exit.\n" << endl;
		if (cinGetOnError)
			cin.get();
		exit(-1);
	}

	NexusMsgStr << "NEXUS data block....found" << endl;

	//  Starts reading the DATA block
	size_t dataBlockIndex = beginBlock + 1;
	while (dataBlockIndex <= endBlock) {
		lineBuf.clear();
		dataOptions.clear();

		//  Reads a line containing an '=' from the DATA block and inserts a ' ' (space) character to the right, so we get '= '
		//  and eliminates the spaces before the '=' character. This helps in tokenizing and identifying the keywords below.
		lineBuf = fileBuf[dataBlockIndex];
		size_t reversePos = lineBuf.rfind('=');
		while (reversePos != string::npos) {
			if ((lineBuf[reversePos + 1] != ' ')
					|| (lineBuf[reversePos + 1] != '\t')
					|| (lineBuf[reversePos + 1] != '\n'))
				lineBuf.insert(reversePos + 1,1,' ');

			pos = reversePos - 1;
			reversePos = lineBuf.length() - reversePos;
			while ((lineBuf[pos] == ' ') || (lineBuf[pos] == '\t') || (lineBuf[pos] == '\n')) {
				lineBuf.erase(pos,1);
				--pos;
			}
			reversePos = lineBuf.rfind('=', lineBuf.length() - reversePos - 1);
		}

		stringstream tokenSStr;
		tokenSStr.str(lineBuf);
		tokenSStr >> dataOptions;
		//  Reads the line "dimensions ntax=xxx nchar=xxx;"
		if (cmp_nocase(dataOptions,"dimensions") == 0) {
			pos = lineBuf.find('=');
			seqNum = seqSize = 0;
			//  Identifies keywords based on the last three letters preceding the '=' character
			while (pos != string::npos) {
				if (cmp_nocase(lineBuf.substr(pos - 3,3),"tax") == 0) {
					stringstream optionSStr;
					optionSStr.str(lineBuf.substr(pos + 1));
					optionSStr >> seqNum;
				}
				else if (cmp_nocase(lineBuf.substr(pos - 3,3),"har") == 0) {
					stringstream optionSStr;
					optionSStr.str(lineBuf.substr(pos + 1));
					optionSStr >> seqSize;
				}
				pos = lineBuf.find('=',pos + 1);
			}
		}
		//  Reads the line "format datatype=xxx interleave=xxx gap=xxx"
		else if (cmp_nocase(dataOptions,"format") == 0) {
			pos = lineBuf.find('=');
			//  Identifies keywords based on the last three letters preceding the '=' character
			while (pos != string::npos) {
				//	keyword = type
				if (cmp_nocase(lineBuf.substr(pos - 3,3),"ype") == 0) {
					stringstream optionSStr;
					optionSStr.str(lineBuf.substr(pos + 1));
					optionSStr >> seqDatatype;
				}
				//	keyword = symbols
				else if (cmp_nocase(lineBuf.substr(pos - 3,3),"ols") == 0) {
					stringstream optionSStr;
					optionSStr.str(lineBuf.substr(pos + 1));
					optionSStr >> seqSymbols;
				}
				//	keyword = gap
				else if (cmp_nocase(lineBuf.substr(pos - 3,3),"gap") == 0) {
					stringstream optionSStr;
					optionSStr.str(lineBuf.substr(pos + 1));
					optionSStr >> seqGapSym;
				}
				//	keyword = missing
				else if (cmp_nocase(lineBuf.substr(pos - 3,3),"ing") == 0) {
					stringstream optionSStr;
					optionSStr.str(lineBuf.substr(pos + 1));
					optionSStr >> seqMisSym;
				}
				pos = lineBuf.find('=',pos + 1);
			}
		}
		//  Reads the data matrix
		else if (cmp_nocase(dataOptions,"matrix") == 0) {
			string dataID, dataSeq;
			tokenSStr >> dataID >> dataSeq;
			while (!tokenSStr.eof()) {
				seqIDVec.push_back(dataID);
				seqVec.push_back(dataSeq);
				tokenSStr >> dataID >> dataSeq;
			}
		}
		++dataBlockIndex;
	}	//  End DATA block read

	//  Checks if the data has been specified as DNA
	if ((seqDatatype != "") && (cmp_nocase(seqDatatype,"dna") != 0) && (cmp_nocase(seqDatatype,"nucleotide") != 0)) {
		cerr << "\nError! The specified \"datatype=" << seqDatatype << "\" option is not supported by Migraine." << endl
			 << "Migraine only accepts a basic DNA format (please consult the help file for examples)."
			 << "\nAborting Migraine...Press any key to exit.\n" << endl;
		if (cinGetOnError)
			cin.get();
		exit(-1);
	}

	//  Checks if the specified dimensions of the data correspond to the actual dimensions of the data matrix
	if (seqNum && seqNum != seqVec.size()) {
		cerr << "\nError! The specified \"ntax\" option does not correspond to the number of sequences." << endl
			 << "Note: Migraine does not support the interleave format for sequences."
			 << "\nAborting Migraine...Press any key to exit.\n" << endl;
		if (cinGetOnError)
			cin.get();
		exit(-1);
	}

	//	if the total number of sequences has not been defined by "ntax" then define it
	if (!seqNum)
		seqNum = seqVec.size();

	//	if the size of the sequences has not been defined by "nchar" then define it
	if (!seqSize)
		seqSize = seqVec[0].length();

	//	initialising the nucleotide states
	for (size_t siteIndex = 0; siteIndex < seqSize; siteIndex++)
		siteState[ (int) siteIndex] = "UN";//CM->RL 23.01.14

	for (size_t i = 0; i < seqNum; i++) {
		//	Checks whether all sequence ID's have been specified
		if (!seqIDVec[i].length()) {
			cerr << "\nError! The ID for sequence " << (i + 1) << " is nonexistent."
				 <<	"\nMigraine needs an ID for every sequence in the NEXUS file: " << fileName
				 << "\nAborting Migraine...Press any key to exit.\n" << endl;
			if (cinGetOnError)
				cin.get();
			exit(-1);
		}

		//  Checks if each sequence of the data matrix corresponds to the specified width of the data matrix
		if (seqSize != seqVec[i].length()) {
			cerr << "\nError! The NEXUS file contains sequences of unequal size."
				 	"\nPlease check the \"nchar\" option or for unaligned sequences or for whitespaces within sequences."
				 << "\nAborting Migraine...Press any key to exit.\n" << endl;
			if (cinGetOnError)
				cin.get();
			exit(-1);
		}
		//  Checks if each sequence of the data matrix contains information other than, gap, missing or the specified datatype
		else {
			for (size_t j = 0; j < seqSize; j++) {
				seqVec[i][j] = toupper(seqVec[i][j]);
				if (!((seqVec[i][j] == 'A')
						|| (seqVec[i][j] == 'T')
						|| (seqVec[i][j] == 'G')
						|| (seqVec[i][j] == 'C')
						|| (seqVec[i][j] == toupper(seqMisSym[0]))
						|| (seqVec[i][j] == toupper(seqGapSym[0])))) {
					
                    if(seqVec[i][j] == '.') {
                        if ( i == 0 || ( ( i == 1 ) && ( cmp_nocase( seqIDVec[0], "anc" ) == 0 ) ) ) {
                            cerr << "\nError! Migraine does not allows to use the (match) character '.' in the first sequence nor for the MRCA (anc)." << endl
                                 << "Migraine only accepts DNA nucleotides (i.e. ATGC) or missing or gap characters for those two sequences."
                                 << "\nAborting Migraine...Press any key to exit.\n" << endl;
                            if ( cinGetOnError )
                                cin.get();
                            exit( -1 );

                        } else {
                            if ( cmp_nocase(seqIDVec[0],"anc") == 0 ) {
                                if ( seqVec[0][j] != seqVec[1][j] ) {
                                    cerr << "\nError! Migraine found the (match) character '.' at position " << j << " in the sequence with ID "
                                         << seqIDVec[i] << "." << endl
                                         << "But nucleotides at this site are difernt in the anc sequence ('" << seqVec[0][j] << "') "
                                         << "and in the first sequence ('" << seqVec[1][j]<< "')."
                                         << "\nAborting Migraine...Press any key to exit.\n" << endl;
                                    if ( cinGetOnError )
                                        cin.get();
                                    exit( -1 );

                                }
                            }
                            seqVec[i][j] = seqVec[0][j];
                        }
                    } else {
                        cerr << "\nError! Migraine couldn't recognize the character \"" << seqVec[i][j] << "\" in the sequence with ID "
                             << seqIDVec[i] << "." << endl
                             << "Migraine only accepts DNA nucleotides (i.e. ATGC) or missing or gap characters."
                             << "\nAborting Migraine...Press any key to exit.\n" << endl;
                        if (cinGetOnError)
                            cin.get();
                        exit(-1);
                    }
				}
				else if ((seqVec[i][j] == toupper(seqMisSym[0])) || (seqVec[i][j] == toupper(seqGapSym[0])))
					siteState[ (int) j] = "NPP";//CM->RL 23.01.14
			}
		}
	}

	//	Checks if the MRCA sequence has been provided as the first sequence of the data block
	if (cmp_nocase(seqIDVec[0],"anc") == 0) {
		seqIDVec.erase(seqIDVec.begin());
		for (size_t it = 0; it < seqVec[0].size(); it++)
			MRCASeq.push_back(seqVec[0][it]);
		seqVec.erase(seqVec.begin());
		--seqNum;
	}
	else
		for (size_t siteIndex = 0; siteIndex < seqSize; siteIndex++)
			MRCASeq.push_back(' ');

	//	counts the (unique) haplotypes in the NEXUS file and checks if the haplotype labels in the NEXUS file are all unique
	size_t initialHapCount = 1;
	for (size_t i = 1; i < seqVec.size(); i++) {
		size_t j = 0, k = 0;
		while ((j != i-1) && (seqVec[j] != seqVec[i]))
			j++;
		if (seqVec[j] != seqVec[i])
			++initialHapCount;

		//	checking for duplicate taxon labels
		while ((k != i - 1) && (seqIDVec[k] != seqIDVec[i]))
			k++;
		if ((k < i - 1) || (seqIDVec[k] == seqIDVec[i])) {
			cerr << "\n(!) Migraine accepts only unique NEXUS taxon labels. Please re-check " << fileName
				 << "\nFound Label: " << seqIDVec[k] << " (line " << k << ")\t" << "and Label: " << seqIDVec[i] <<  " (line " << i << ")\t" << endl
				 << "\nAborting Migraine...Press any key to exit.\n" << endl;
			if (cinGetOnError)
				cin.get();
			exit(-1);
		}
	}

	size_t iniSeqSize = 0;
	for (size_t j = 0; j < seqSize; j++)
		if (siteState[ (int) j] != "NPP")//CM->RL 23.01.14
			++iniSeqSize;

	hapCount = initialHapCount;
	samSize = gpSample->Size;
	NexusMsgStr << "\nTotal sequence length (with gaps and missing data) : " << seqSize << endl
		 << "\nInitial sequence length (WITHOUT gaps  and missing data) : " << iniSeqSize << endl
		 << "Initial number of haplotypes (excluding duplicate sequences) : " << hapCount << endl
		 << "\nInitial sample size : " << gpSample->Size << " individual(s)" << endl;
}


void CNexus::FillHaplotypeInfoFromSNP(CSample *gpSample)
{
	//	defines the total number of sequences = always 3 here (minimal choice = ancstate/ancestral/derived)
	seqNum = 3;

	//	define the size of the sequences, here it is always 1
    seqSize = 1;

	//	initialising the nucleotide states RL->CBR pourquoi?
    siteState[0] = "UN";//CM->RL 23.01.14

    //Fill ID for 2 haplotypes 0/1 : ancestral state 001 and derived state 001 : these are the names of the alleles in the genepop file. RL->CBR
    seqIDVec.resize(3);
    seqIDVec[0]="Anc";seqIDVec[1]="001";seqIDVec[2]="002";

    //Fill seq pour 2 haplotypes : ancestral state 001 and derived state 001 : these are the names of the alleles in the genepop file. RL->CBR
    seqVec.resize(3);
    seqVec[0]="A";seqVec[1]="A";seqVec[2]="T";

    //	Checks if the MRCA sequence has been provided as the first sequence of the data block
	if (cmp_nocase(seqIDVec[0],"anc") == 0) {
		seqIDVec.erase(seqIDVec.begin());
		for (size_t it = 0; it < seqVec[0].size(); it++)
			MRCASeq.push_back(seqVec[0][it]);
		seqVec.erase(seqVec.begin());
		--seqNum;
	}

	//	counts the (unique) haplotypes in the NEXUS file and checks if the haplotype labels in the NEXUS file are all unique
	size_t initialHapCount = 1;
	for (size_t i = 1; i < seqVec.size(); i++) {
		size_t j = 0, k = 0;
		while ((j != i-1) && (seqVec[j] != seqVec[i]))
			j++;
		if (seqVec[j] != seqVec[i])
			++initialHapCount;

		//	checking for duplicate taxon labels
		while ((k != i - 1) && (seqIDVec[k] != seqIDVec[i]))
			k++;
		if ((k < i - 1) || (seqIDVec[k] == seqIDVec[i])) {
			cerr << "\n(!) Migraine accepts only unique NEXUS taxon labels. Please re-check " << fileName
            << "\nFound Label: " << seqIDVec[k] << " (line " << k << ")\t" << "and Label: " << seqIDVec[i] <<  " (line " << i << ")\t" << endl
            << "\nAborting Migraine...Press any key to exit.\n" << endl;
			if (cinGetOnError)
				cin.get();
			exit(-1);
		}
	}

	size_t iniSeqSize = 0;
	for (size_t j = 0; j < seqSize; j++)
		if (siteState[ (int) j] != "NPP")//CM->RL 23.01.14
			++iniSeqSize;

	hapCount = initialHapCount;
	samSize = gpSample->Size;
	NexusMsgStr << "\nTotal sequence length (with gaps and missing data) : " << seqSize << endl
    << "\nInitial sequence length (WITHOUT gaps  and missing data) : " << iniSeqSize << endl
    << "Initial number of haplotypes (excluding duplicate sequences) : " << hapCount << endl
    << "\nInitial sample size : " << gpSample->Size << " individual(s)" << endl;
}

void CNexus::removeNEXUSComments()
{
/*
 	 Removes NEXUS comments from a string as long as they are delimited by the square brackets []
 */

	size_t beginComment, endComment;
	bool commentFound = false;
	beginComment = endComment = string::npos;
	beginComment = lineBuf.find('[');
	endComment = lineBuf.find(']');

	if ((beginComment != string::npos) && (endComment != string::npos))
		commentFound = true;
	else if (((beginComment != string::npos) && (endComment == string::npos))
				|| ((beginComment == string::npos) && (endComment != string::npos))) {
		cerr << "\nError! NEXUS comment block started/ended unexpectedly."
			 << "\nAborting Migraine...Press any key to exit.\n" << endl;
		if (cinGetOnError)
			cin.get();
		exit(-1);
	}

	while (commentFound) {
		lineBuf.erase(beginComment,endComment - beginComment + 1);

		commentFound = false;
		beginComment = endComment = string::npos;
		beginComment = lineBuf.find('[');
		endComment = lineBuf.find(']');

		if ((beginComment != string::npos) && (endComment != string::npos))
			commentFound = true;
		else if (((beginComment != string::npos) && (endComment == string::npos))
					|| ((beginComment == string::npos) && (endComment != string::npos))) {
			cerr << "\nError! NEXUS comment block started/ended unexpectedly."
				 << "\nAborting Migraine...Press any key to exit.\n" << endl;
			if (cinGetOnError)
				cin.get();
			exit(-1);
		}
	}
}

void CNexus::checkIfDataISMCompatible(CSample *gpSample)
{
/*
	Checks if the sequence data extracted from the NEXUS file is compatible with the basic ISM model assumptions
	Among other tasks, it checks if a perfect phylogeny exists and in some cases modifies data to make it possible.
*/

	//	before eliminating duplicate genes outlier counters are initialised and mapped to itself
	bool checkForOutlierGenes;
	if (outlierGeneFileBool && !promptHyperSegSites) {
		for (size_t i = 0; i < seqNum; i++)
			outlierGeneCount[seqIDVec[i]] = vector<int> (2,0);
		checkForOutlierGenes = true;
	}

	//	removes duplicates when the Nexus file contains sequences from all individuals
	removeDuplicates();

    // Establishes a link between the allelic states/labels from the Genepop file and the corresponding NEXUS taxon labels
    mapNexusToGenepop(gpSample,false);
	NexusMsgStr << "\nMapping NEXUS data and Genepop data....done" << endl;

	origInfSites = currInfSites;

	NexusMsgStr << "\nNumber of informative sites : " << currInfSites << endl
		 << "Number of remaining haplotypes : " << seqNum << endl;

	// sets the MRCA/reference sequence if nothing has been specified by the user
	setMRCA();

	//	handles nucleotides displaying more than two alleles
	processHyperSegSites(gpSample);

	bool resolve4GTBool = false;
	do {
		seqBinMat.clear();
		fourGameteTestMat.clear();
		modHapMatSiteIndex.clear();
		char userChoice = ' ';

		//	remove duplicate sequences IF necessary and after the removal of incompatible sites
		mergeDupHapCounts(gpSample, true);

		//  converts compatible nucleotides into a 0,1 format
		seqBinMat.resize(seqNum);
		for (size_t i = 0; i < seqNum; i++)
			for (size_t j = 0; j < seqSize; j++)
				if (siteState[ (int) j] == "CPP") {//CM->RL 23.01.14
					if (seqVec[i][j] != MRCASeq[j])
						seqBinMat[i].push_back(1);
					else
						seqBinMat[i].push_back(0);
					modHapMatSiteIndex.push_back( (int) j);//CM->RL 23.01.14
				}

		fourGameteTestMat.resize(seqBinMat[0].size());
		for (size_t j = 0; j < seqBinMat[0].size(); j++)
			fourGameteTestMat[j].resize(seqBinMat[0].size(), " ");

		//	performs the four gamete test for all pairs of nucleotides
		checkForPerfectPhylogeny();

		//	counts the number of 4GT incompatibilities for every gene
		if (outlierGeneFileBool && !promptHyperSegSites && checkForOutlierGenes) {
			checkForOutlierGenes = false;
			for (size_t i = 0; i < fourGameteTestPerHap.size(); i++) {
				if (mapHap2Genes.find(seqIDVec[i]) == mapHap2Genes.end()) {
					cerr << "\nError in void CNexus::checkIfDataISMCompatible "
							"\nmapHap2Genes.find(seqIDVec[i]) == mapHap2Genes.end()"
						 <<	"\nCould not find the sequence : " << seqIDVec[i]
						 << "\n Please contact the authors."
							"\nAborting Migraine...Press any key to exit.\n";
					if (cinGetOnError)
						cin.get();
					exit(-1);
				}
				else
					for (vector<string>::iterator it = mapHap2Genes[seqIDVec[i]].begin(); it != mapHap2Genes[seqIDVec[i]].end(); ++it)
						outlierGeneCount[*it][1] += fourGameteTestPerHap[i];
			}

			stringstream stst;
			stst << "OutlierGenes_" << locNum << ".txt";
	        ofstream outlierGenes (stst.str().c_str(), std::ios::out);

//	    	outlierGenes << "Seq_ID\tMinor_allele\t4GT\tTotal" << endl;
//	        for (map<string, vector<int> >::iterator it = outlierGeneCount.begin(); it != outlierGeneCount.end(); ++it)
//	        	outlierGenes << (*it).first << "\t"
//	        				 << (*it).second[0] << "\t"
//	        				 << (*it).second[1] << "\t"
//	        				 << ((*it).second[0] + (*it).second[1]) << endl;

	    	outlierGenes << "Seq_ID\tMult\tMin\t4GT\tTotal" << endl;
	        for (size_t i = 0; i < seqNum; i++)
	        	outlierGenes << seqIDVec[i] << "\t"
	        				 << mapHap2Genes[seqIDVec[i]].size() << "\t"
	        				 << outlierGeneCount[seqIDVec[i]][0] << "\t"
	        				 << outlierGeneCount[seqIDVec[i]][1] << "\t"
	        				 << (outlierGeneCount[seqIDVec[i]][0] + outlierGeneCount[seqIDVec[i]][1]) << endl;
	        outlierGenes.close();
	        exit(-1);
		}

		if (!pass4GTBool) {
			resolve4GTBool = true;
			printHaps(false);
            stringstream tempStrstr;

			int siteMaxNum = *max_element(fourGameteTestCount.begin(),fourGameteTestCount.end()),
				seqMaxNum = *max_element(fourGameteTestPerHap.begin(),fourGameteTestPerHap.end());
			size_t eraseSiteIndex = modHapMatSiteIndex[max_element(fourGameteTestCount.begin(),fourGameteTestCount.end()) - fourGameteTestCount.begin()],
				   eraseSeqIndex = max_element(fourGameteTestPerHap.begin(),fourGameteTestPerHap.end()) - fourGameteTestPerHap.begin();

            tempStrstr << "";
            tempStrstr << "\nMaximum number of \"X\" counts contained at a single nucleotide site..."
					"\n\tSite/Nucleotide number : " << eraseSiteIndex + 1 << " (" << siteMaxNum << " incompatibilities)."
					"\n\nMaximum number of \"X\" counts contained in a single haplotype..."
					"\n\tHaplotype : " << seqIDVec[eraseSeqIndex] << " (" << seqMaxNum << " incompatibilities), which corresponds to "
				 << gpSample->StateEffectif[hapAlleleIndex[seqIDVec[eraseSeqIndex]]] << " individual(s)." << endl;
            NexusMsgStr << tempStrstr.str();
            cout << tempStrstr.str();
			if (cmp_nocase(resolveForPerfectPhylogeny, "manual") == 0) {
                stringstream tempStrstr;
                tempStrstr << "\nN.B. Eliminating the site numbered " << eraseSiteIndex + 1 << " NECESSARILY resolves all of its incompatibilities from the "
						"\n4GT matrix but eliminating a haplotype " << seqIDVec[eraseSeqIndex] << " DOESN'T NECESSARILY resolve all of its incompatibilities." << endl;

				tempStrstr << "\nWould you like to eliminate "
						"\nthe specified (N)ucleotide, the (H)aplotype or a(B)ort Migraine ?"
						"\nPlease enter a single case-sensitive character...";
                NexusMsgStr << tempStrstr.str();
                cout << tempStrstr.str();
				while (true) {	//	*** infinite loop ***
					cin >> userChoice;
					cin.ignore();
                    if ((userChoice != 'N') && (userChoice != 'H') && (userChoice != 'B')) {
                        tempStrstr << "";
						tempStrstr << "\n(!) Migraine did not understand your choice"
								"\nWould you like to eliminate "
								"\nthe specified (N)ucleotide, the (H)aplotype or a(B)ort Migraine ?"
								"\nPlease enter a single case-sensitive character...";
                        NexusMsgStr << tempStrstr.str();
                        cout << tempStrstr.str();
                    } else
						break;
				};
				if (userChoice == 'N') {
					//	marking the nucleotide site as incompatible, will be ignored in the next loop
                    tempStrstr << "";
                    tempStrstr << "\nEliminating site number : " << eraseSiteIndex + 1 << " for all sequences...\n";
                    NexusMsgStr << tempStrstr.str();
                    cout << tempStrstr.str();

					siteState[ (int) eraseSiteIndex] = "NPP";//CM->RL 23.01.14
					mergeDupHapCounts(gpSample, false);
				}
				else if (userChoice == 'H') {
					//	eliminating the haplotype contributing to the maximum number of X counts/ISM incompatibilities
                    tempStrstr << "";
                    tempStrstr << "\neliminating the haplotype '" << seqIDVec[eraseSeqIndex] << "' contributing to the maximum number of X counts/ISM incompatibilities\n";
                    NexusMsgStr << tempStrstr.str();
                    cout << tempStrstr.str();
					removeHap(gpSample, (int) eraseSeqIndex,seqIDVec[eraseSeqIndex],false);//CM->RL 23.01.14
				}
				else {
					cerr << "\nAborting Migraine...Press any key to exit.\n";
					if (cinGetOnError)
						cin.get();
					exit(-1);
				}
			}
			else if (cmp_nocase(resolveForPerfectPhylogeny, "sites") == 0) {
				//	Autoamtically marking the nucleotide site as incompatible, will be ignored in the next loop
				NexusMsgStr << "\nAUTOMATICALLY eliminating site numbered " << eraseSiteIndex + 1 << " for all sequences...\n";
				siteState[ (int) eraseSiteIndex] = "NPP";//CM->RL 23.01.14
				mergeDupHapCounts(gpSample, false);
			}
			else if (cmp_nocase(resolveForPerfectPhylogeny, "haps") == 0) {
				//	Automatically eliminating the haplotype contributing to the maximum number of X counts/ISM incompatibilities
				NexusMsgStr << "\nAUTOMATICALLY eliminating the haplotype " << seqIDVec[eraseSeqIndex] << "...";
				removeHap(gpSample, (int) eraseSeqIndex,seqIDVec[eraseSeqIndex],false);//CM->RL 23.01.14
			}
			else if (cmp_nocase(resolveForPerfectPhylogeny, "auto") == 0) {
				if (siteMaxNum >= seqMaxNum) {
					//	Autoamtically marking the nucleotide site as incompatible, will be ignored in the next loop
					NexusMsgStr << "\nAUTOMATICALLY eliminating site numbered " << eraseSiteIndex + 1 << " for all sequences...\n";
					siteState[ (int) eraseSiteIndex] = "NPP";//CM->RL 23.01.14
					mergeDupHapCounts(gpSample, false);
				}
				else {
					//	Automatically eliminating the haplotype contributing to the maximum number of X counts/ISM incompatibilities
					NexusMsgStr << "\nAUTOMATICALLY eliminating the haplotype " << seqIDVec[eraseSeqIndex] << "...";
					removeHap(gpSample, (int) eraseSeqIndex,seqIDVec[eraseSeqIndex],false);//CM->RL 23.01.14
				}
			}
			else {
				cerr << "\nPlease specify the keyword \"ResolveForPerfectPhylogeny\" in the settings file in order to proceed."
					 << "\nAborting Migraine...Press any key to exit.\n";
				if (cinGetOnError)
					cin.get();
				exit(-1);
			}
		}
	} while (!pass4GTBool);

	countNucStates();

	if (recodeISMBool) {
		cerr << "\nWarning! A perfect phylogeny is possible for the haplotype data in \"" << fileName << "\""
			 << "\nbut the ancestral and derived states need to be swapped for some of the segregating sites."
//			 << "\nMigraine can take care of this swap automatically. Press any key to continue...\n";
			 << "\nWhile Migraine can take care of this swap automatically, this option is presently disabled."
			 << "\nPlease contact the authors."
			 	"\nPress any key to exit.\n";

		printHaps(true);
		cerr << "\nFinal number of informative sites : " << currInfSites << " (out of " << origInfSites << " sites)" << endl
			 << "Final number of haplotypes : " << seqNum << " (out of " << hapCount << " haplotypes)" << endl;
		cerr << "\nFinal sample size : " << gpSample->Size << " (out of " << samSize << " individuals)" << endl << endl;

		if (cinGetOnError)
			cin.get();

		exit(-1);

		recodeISMSeq();
	}

	//	get the updated size of the sequences (i.e. not containing ISM incompatibilities)
	size_t newSeqSize = 0;
	for (size_t siteIndex = 0; siteIndex < seqSize; siteIndex++)
		if (siteState[ (int) siteIndex] == "CPP")//CM->RL 23.01.14
			++newSeqSize;

	//  converts the haplotype matrix into a tree format (or mutation path to the MRCA)
	mutPathMatrix.resize(seqNum);
	siteMutCountVec.clear();
	siteMutCountVec.assign(newSeqSize, 0);
	for (size_t i = 0; i < seqNum; i++) {
		for (size_t j = 0; j < newSeqSize; j++)
			if (seqBinMat[i][j] == 1) {
				mutPathMatrix[i].push_back( (int) (j + 1) );//CM->RL 23.01.14
				++siteMutCountVec[j];
			}
		mutPathMatrix[i].push_back(0);
	}

    stringstream tempStrstr;
    tempStrstr << "";

    if (resolve4GTBool)
        tempStrstr << "\nAll incompatibilities have been successfully resolved!" << endl;
        
	printHaps(true);
	tempStrstr << "\nFinal number of informative sites : " << currInfSites << " (out of " << origInfSites << " sites)" << endl
		 << "Final number of haplotypes : " << seqNum << " (out of " << hapCount << " haplotypes)" << endl;
	tempStrstr << "\nFinal sample size : " << gpSample->Size << " (out of " << samSize << " individuals)" << endl << endl;
    NexusMsgStr << tempStrstr.str();
    cout << tempStrstr.str();


	if (DebugNexusSample) {
		NexusMsgStr << "\nmutPathMatrix = \n";
		for (size_t i = 0; i < seqNum; i++) {
			for (size_t j = 0; j < mutPathMatrix[i].size(); j++)
				NexusMsgStr << mutPathMatrix[i][j] << " ";
			NexusMsgStr << "\n";
		}
		NexusMsgStr << "siteMutCountVec = \n";
		for (size_t j = 0; j < siteMutCountVec.size(); j++)
			NexusMsgStr << siteMutCountVec[j] << " ";
		NexusMsgStr << "\n\n\n";
	}

    if (HobolthISM) {
  	  DmVec.assign(siteMutCountVec.size(),0);
  	  for (int k = 0; k < gpSample->Sam_AlleleNbr; k++)
  		  for (vector<int>::iterator it = mutPathMatrix[k].begin(); *it != 0; it++)
  			  DmVec[*it - 1] += gpSample->StateEffectif[k];
    }

    gpSample->mutPathMatrix = mutPathMatrix;
	gpSample->siteMutCountVec = siteMutCountVec;
	gpSample->DmVec = DmVec;
    CopyCatNexus.copyCSample(gpSample);

	ofstream outFile (outFilename.c_str(), std::ios::out);
	outFile << NexusMsgStr.str();
	outFile.close();
}

void CNexus::removeDuplicates()
{
/*
	identifies and filters out the duplicate sequences obtained from the NEXUS file
*/

	//	counts the (unique) haplotypes in the NEXUS file and identifies the duplicates
	vector<int> dupHapIndex, origHapIndex;
	size_t dupHapCount = 0;

	//	identify the duplicate sequences
	mapHap2Genes[seqIDVec[0]].push_back(seqIDVec[0]);
	for (size_t i = 1; i < seqVec.size(); i++) {
		size_t j = 0;
		while ((j != i-1) && (seqVec[j] != seqVec[i]))
			j++;
		if ((j < i - 1) || (seqVec[j] == seqVec[i])) {
			dupHapIndex.push_back( (int) i);//CM->RL 23.01.14
			origHapIndex.push_back( (int) j);//CM->RL 23.01.14
			//	updates for keeping count of the number of minor allele states for every gene
			if (outlierGeneFileBool && !promptHyperSegSites)
				mapHap2Genes[seqIDVec[j]].push_back(seqIDVec[i]);
			++dupHapCount;
		}

		if (outlierGeneFileBool && !promptHyperSegSites && (seqVec[j] != seqVec[i]))
			mapHap2Genes[seqIDVec[i]].push_back(seqIDVec[i]);
	}

	//	remove the duplicate haplotypes
	if (dupHapIndex.size() > 0) {
		NexusMsgStr << "\nFound " << dupHapCount << " duplicate sequence(s) in the Nexus file " << fileName << ". \nEliminating them..." << endl;

		//	reverse elimination of seqIDVec and seqVec in order to preserve the order of duplicate haplotype indices
		for (int dups =  ( (int) dupHapIndex.size() ) - 1; dups >= 0; dups--) { //CM->RL 23.01.14
			if (DebugNexusSample)
				NexusMsgStr << seqIDVec[dupHapIndex[dups]] << " -> " << seqIDVec[origHapIndex[dups]] << endl;

			//	updates for keeping count of the number of minor allele states for every gene
			if (outlierGeneFileBool && !promptHyperSegSites)
				mapHap2Genes[seqIDVec[origHapIndex[dups]]].push_back(seqIDVec[dupHapIndex[dups]]);

			seqIDVec.erase(find(seqIDVec.begin(),seqIDVec.end(),seqIDVec[dupHapIndex[dups]]));
			vector<string> delStr(1,seqVec[dupHapIndex[dups]]);
			seqVec.erase(find_end(seqVec.begin(),seqVec.end(),delStr.begin(),delStr.end()));
		}
		seqNum = seqVec.size();
	}
}

void CNexus::mapNexusToGenepop(CSample *gpSample, bool update)
{
/*
	Checks whether the allelic states/labels from the Genepop file correspond to the NEXUS taxon labels/haplotype order
*/

	if (nexusNumberLabels) {
		int taxLabel;
		//	element-wise check to see if the NUMERIC taxon labels in the Nexus file correspond to the Genepop file ALLELE INDICES
		for (size_t seqIndex = 0; seqIndex < seqNum; seqIndex++) {
			//	transforming strings into int for an easier comparison of the taxon labels
			stringstream stst(seqIDVec[seqIndex]);
			stst >> taxLabel;

			vector<int>::iterator it = find(gpSample->alleleIndices.begin(), gpSample->alleleIndices.end(), taxLabel);
			if (it == gpSample->alleleIndices.end()) {
				if (!update) {
					cerr << "\n(!) Could not find the Nexus taxon label " << seqIDVec[seqIndex]
					     << " as an allele index in the corresponding Genepop file at locus number "
						 << locNum << "(" << ProcessedLoci[locNum - 1].locusName << ")"
						 << "\nPlease cross-check the Genepop file : " << ProcessedLoci[locNum - 1].fileName << "\nand the Nexus file : " << fileName << "\n"
						 << "\nAborting Migraine...Press any key to exit.\n" << endl;
				}
				else {
					cerr << "\n(!) Error in CNexus::mapNexusToGenepop : Could not find the Nexus taxon label " << seqIDVec[seqIndex]
					     << " as an allele index in gpSample->alleleIndices.end() at locus number "
						 << locNum << "(" << ProcessedLoci[locNum - 1].locusName << ")"
						 << "\nPlease contact the authors.\n"
						 << "\nAborting Migraine...Press any key to exit.\n" << endl;
				}
				if (cinGetOnError)
					cin.get();
				exit(-1);
			}
			else
				hapAlleleIndex[seqIDVec[seqIndex]] = (int) ( it - gpSample->alleleIndices.begin());
		}
	}
	else {
		//	element-wise check to see if the SPECIFIC taxon labels in the Nexus file correspond to the Genepop file INDIVIDUAL LABELS
		for (size_t seqIndex = 0; seqIndex < seqNum; seqIndex++) {
			//	CBR : NB. Genepop parses the individual labels with a single space at the end
			map<string, int>::iterator it = ProcessedLoci[locNum - 1].mapIndLab2State.find(seqIDVec[seqIndex] + " ");
			if (it == ProcessedLoci[locNum - 1].mapIndLab2State.end()) {
				if (!update) {
					cerr << "\n(!) Could not find the Nexus taxon label " << seqIDVec[seqIndex]
						 << " as an individual label in the corresponding Genepop file at locus number "
						 << locNum << "(" << ProcessedLoci[locNum - 1].locusName << ")"
						 << "\nPlease cross-check the Genepop file : " << ProcessedLoci[locNum - 1].fileName << "\nand the Nexus file : " << fileName << "\n"
						 << "\nAborting Migraine...Press any key to exit.\n" << endl;
				}
				else {
					cerr << "\n(!) Error in CNexus::mapNexusToGenepop : Could not find the Nexus taxon label " << seqIDVec[seqIndex]
					     << " as an allele index in ProcessedLoci[locNum - 1].mapIndLab2State.end() at locus number "
						 << locNum << "(" << ProcessedLoci[locNum - 1].locusName << ")"
						 << "\nPlease contact the authors.\n"
						 << "\nAborting Migraine...Press any key to exit.\n" << endl;
				}
				if (cinGetOnError)
					cin.get();
				exit(-1);
			}
			else {
				int typage = ProcessedLoci[locNum - 1].mapIndLab2State[seqIDVec[seqIndex] + " "];
				vector<int>::iterator it = find(gpSample->alleleIndices.begin(), gpSample->alleleIndices.end(), typage);
				hapAlleleIndex[seqIDVec[seqIndex]] = (int) ( it - gpSample->alleleIndices.begin());
			}
		}
	}

	//	refresh nucleotides states
	countNucStates();
}

void CNexus::countNucStates()
{
/*
	identifies nucleotides displaying no alleles or more than two alleles (incompatible with the ISM assumptions)
*/

	//	checks for non-segregating and hyper-segregating (w.r.t. to the ISM model) sites
	siteMutCountVec.clear();
	siteMutCountVec.assign(seqSize, 0);
	size_t nonSegCount = 0, otherCount = 0;
	hyperSegCount = 0;
	for (size_t siteIndex = 0; siteIndex < seqSize; siteIndex++) {
		if (siteState[ (int) siteIndex] != "NPP") {//CM->RL 23.01.14
			map<char, int> nucFreqCount;
			int siteSegCount = 1;

			//	counts for A,T,G,C at each nucleotide position
			++nucFreqCount[seqVec[0][siteIndex]];

			for (size_t i = 1; i < seqNum; i++) {
				size_t j = 0;
				while ((j != i-1) && (seqVec[j][siteIndex] != seqVec[i][siteIndex]))
					j++;
				if (seqVec[j][siteIndex] != seqVec[i][siteIndex])
					siteSegCount++;

				//	counts for A,T,G,C at each nucleotide position
				++nucFreqCount[seqVec[i][siteIndex]];
			}
			if (siteSegCount == 1) {
				++nonSegCount;
				siteState[ (int) siteIndex] = "NS"; //CM->RL 23.01.14
			}
			else if (siteSegCount > 2) {
				++hyperSegCount;
				siteState[ (int) siteIndex] = "HS"; //CM->RL 23.01.14
			}
			else
				siteState[ (int) siteIndex] = "CPP";//CM->RL 23.01.14

			siteMutCountVec[siteIndex] = siteSegCount;
			siteStatesCount[ (int) siteIndex] = nucFreqCount;//CM->RL 23.01.14
		}
		else if (siteState[ (int) siteIndex] == "NS")//CM->RL 23.01.14
			++nonSegCount;
		else if (siteState[ (int) siteIndex] == "NPP")//CM->RL 23.01.14
			++otherCount;
	}

	if (currInfSites != (seqSize - nonSegCount - otherCount))
		currInfSites = seqSize - nonSegCount - otherCount;
}

void CNexus::setMRCA()
{
/*
	sets the MRCA/reference sequence by choosing the most frequent allele at each nucleotide site
*/

	//	if no MRCA has been specified then set the MRCA states
	bool dispOnlyOnce = false;
	for (size_t siteIndex = 0; siteIndex < seqSize; siteIndex++) {
		//	this means that the MRCA has not been specified in the Nexus file
		//	excluding ISM incompatible sites (but includes "NS" and "HS" states)
		if ((MRCASeq[siteIndex] == ' ') && (siteState[ (int) siteIndex] != "NPP")) {//CM->RL 23.01.14
			dispOnlyOnce = true;
			char max = siteStatesCount[ (int) siteIndex].begin()->first;//CM->RL 23.01.14
			map<char, int>::iterator pos;

			for (pos = siteStatesCount[ (int) siteIndex].begin(); pos != siteStatesCount[ (int) siteIndex].end(); ++pos) //CM->RL 23.01.14 *2
				if (pos->second > siteStatesCount[ (int) siteIndex][max])//CM->RL 23.01.14
					max = pos->first;

			MRCASeq[siteIndex] = max;
		}
	}

	if (dispOnlyOnce) {
		NexusMsgStr << "\nThe MRCA/reference sequence was not specified in the NEXUS file."
			 << "\nThe MRCA sequence for ISM compatible sites has been set to : \n";
	}
	else
		NexusMsgStr << "\nThe MRCA sequence for ISM compatible sites has been provided by the user : \n";

	for (size_t siteIndex = 0; siteIndex < seqSize; siteIndex++)
		if (siteState[ (int) siteIndex] == "CPP")//CM->RL 23.01.14
			NexusMsgStr << MRCASeq[siteIndex];
	NexusMsgStr << endl;
}

void CNexus::processHyperSegSites(CSample *gpSample)
{
/*
	prompts the user for a decision on sites displaying more than two alleles
*/

	//	IF hyper-segregating sites are found (w.r.t. to the ISM model)
	if (hyperSegCount > 0) {
		char userChoice = ' ';
        stringstream tempStrstr;
        tempStrstr << "";

		tempStrstr << "\nA total of " << hyperSegCount << " site(s) in \"" << fileName << "\" have more than two allelic states\n";

		for (size_t siteIndex = 0; siteIndex < seqSize; siteIndex++) {
			if (siteState[ (int) siteIndex] == "HS") {//CM->RL 23.01.14
				//	loop as long as the number of allelic states at this site is greater than 2
				do {
					tempStrstr << "\nAllelic counts at site number : " << siteIndex + 1 << endl;
					for (map<char, int>::iterator it = siteStatesCount[ (int) siteIndex].begin(); it != siteStatesCount[ (int) siteIndex].end(); ++it) {//CM->RL 23.01.14 *2
						//	counts the number of individuals for a given allelic state at this nucleotide
						int stateIndCount = 0;
						for (size_t seqIndex = 0; seqIndex < seqNum; seqIndex++)
							if (seqVec[seqIndex][siteIndex] == it->first)
								stateIndCount += gpSample->StateEffectif[hapAlleleIndex[seqIDVec[seqIndex]]];
						tempStrstr << "\t" << it->first << " : " << it->second << "\tpresent among " << stateIndCount << " individual(s)" << endl;
					}
					//	prompt the user for quitting the program, eliminating the individuals or the nucleotide site for all sequences
					if (promptHyperSegSites) {
						tempStrstr << "Do you want to a(B)ort Migraine, (E)liminate this site or eliminate the individuals associated "
								"\nwith any one of the nucleotide states displayed above [e.g. (A),(T),(G),(C)] ?"
								"\nPlease enter a single case-sensitive character...";
                        NexusMsgStr << tempStrstr.str();
                        cout << tempStrstr.str();

						while (true) {	//	*** infinite loop ***
							cin >> userChoice;
							cin.ignore();
							if ((siteStatesCount[ (int) siteIndex].find(userChoice) == siteStatesCount[ (int) siteIndex].end())
                                && (userChoice != 'B') && (userChoice != 'E')) {//CM->RL 23.01.14
                                tempStrstr << "";
								tempStrstr << "\nMigraine did not understand your choice/could not find the specified nucleotide state at this site..."
									 << "\nDo you want to a(B)ort Migraine, (E)liminate this site or eliminate the individuals associated "
										"\nwith any one of the nucleotide states displayed above [e.g. (A),(T),(G),(C)] ?"
										"\nPlease enter a single case-sensitive character...";
                            NexusMsgStr << tempStrstr.str();
                            cout << tempStrstr.str();

                        } else
								break;
						};
						if (userChoice == 'B') {
							cerr << "\nAborting Migraine...Press any key to exit.\n";
							if (cinGetOnError)
								cin.get();
							exit(-1);
						}
						else if (userChoice == 'E') {
                            tempStrstr << "";
							tempStrstr << "\nEliminating site number : " << siteIndex + 1 << " for all sequences...\n";
                            NexusMsgStr << tempStrstr.str();
                            cout << tempStrstr.str();

							siteState[ (int) siteIndex] = "NPP";//CM->RL 23.01.14
							mergeDupHapCounts(gpSample, false);
							break;
						}
						else {
                            tempStrstr << "";
							//	reverse elimination of the haplotypes with the chosen allele
							tempStrstr << "\nEliminating the following sequences associated with allele " << userChoice << "...\n";
                            NexusMsgStr << tempStrstr.str();
                            cout << tempStrstr.str();

							for (int seqIndex =  ( (int) seqNum ) - 1; seqIndex >= 0; seqIndex--) //CM->RL 23.01.14
								if (seqVec[seqIndex][siteIndex] == userChoice)
									removeHap(gpSample,seqIndex,seqIDVec[seqIndex],false);
							siteStatesCount[ (int) siteIndex].erase(userChoice);//CM->RL 23.01.14
						}
					}
					//	by default eliminate this hyper-segregating site for all haplotypes
					else {
                        tempStrstr << "";
						tempStrstr << "Automatically eliminating the nucleotide site/position number : " << siteIndex + 1 << " for all sequences...\n"
                        << "(this is thye default behavior, see key word 'PromptForHyperSegSites' for more options.)" << endl;
                        NexusMsgStr << tempStrstr.str();
                        cout << tempStrstr.str();

						siteState[ (int) siteIndex] = "NPP";//CM->RL 23.01.14

						//	counts the number of minor allele states for every gene
						for (size_t seqIndex = 0; seqIndex < seqNum; seqIndex++) {
							string seqID = seqIDVec[seqIndex];
							//	if this is a minor allele site and a haplotype sequence (unique gene)
							if ((seqVec[seqIndex][siteIndex] != MRCASeq[siteIndex]) &&
									(mapHap2Genes.find(seqID) != mapHap2Genes.end()))
								for (vector<string>::iterator it = mapHap2Genes[seqID].begin(); it != mapHap2Genes[seqID].end(); ++it)
									++outlierGeneCount[*it][0];
						}

						mergeDupHapCounts(gpSample, true);
						break;
					}
				} while (siteMutCountVec[siteIndex] > 2);
			}
		}
	}
}

void CNexus::mergeDupHapCounts(CSample *gpSample, bool outlierBool)
{
/*
	identifies duplicate sequences and adjusts its allele frequencies (from the Genepop file)
*/

	//	counts the (unique) haplotypes in the NEXUS file and identifies the duplicates
	//	by skipping the non-candidate sites (cf. CNexus::siteState)
	vector<int> dupHapIndex, origHapIndex;
	size_t reCalcHapCount = 1, dupHapCount = 0;

	//	initialise modSeqVec with sequences not containing ISM incompatibilities
	modSeqVec.clear();
	for (size_t i = 0; i < seqNum; i++) {
		string seq;
		for (size_t siteIndex = 0; siteIndex < seqSize; siteIndex++)
			if (siteState[ (int) siteIndex] == "CPP")//CM->RL 23.01.14
				seq += seqVec[i][siteIndex];
		modSeqVec.push_back(seq);
	}

	//	identifying the duplicate sequences
	for (size_t i = 1; i < modSeqVec.size(); i++) {
		size_t j = 0;
		while ((j != i-1) && (modSeqVec[j] != modSeqVec[i]))
			j++;
		if (modSeqVec[j] != modSeqVec[i])
			++reCalcHapCount;
		if ((j < i - 1) || (modSeqVec[j] == modSeqVec[i])) {
			dupHapIndex.push_back( (int) i);//CM->RL 23.01.14
			origHapIndex.push_back( (int) j);//CM->RL 23.01.14
			++dupHapCount;
		}
	}

	if (reCalcHapCount != seqVec.size())
		NexusMsgStr << "\nDue to the previous modification(s), Migraine found " << dupHapCount
			 << " duplicate sequence(s). \nEliminating the duplicate(s) and adjusting the Genepop frequencies accordingly..." << endl;

	//	reverse indexing of seqIDVec and seqVec in order to preserve the order of duplicate haplotype indices
	//	during their elimination
	for (int dups =  ( (int) dupHapIndex.size() ) - 1; dups >= 0; dups--) { //CM->RL 23.01.14
		int augmentedHapSize = 0;
		//	moving (pop-wise) all the duplicate haplotype frequencies to the "first of its kind" haplotype
		for (size_t pop = 0; pop < gpSample->Effectif.size(); pop++) {
			//	using pointers in order to augment code readability
			int *fromHapFreq = &(gpSample->Effectif[pop][hapAlleleIndex[seqIDVec[dupHapIndex[dups]]]]);
			int *toHapFreq = &(gpSample->Effectif[pop][hapAlleleIndex[seqIDVec[origHapIndex[dups]]]]);
			*toHapFreq += *fromHapFreq;

			//	keep count of the "first of its kind" haplotype
			augmentedHapSize += *toHapFreq;
		}

		//	check if the modified total (i.e. across pops) allelic counts are consistent with the operation above
		int *totFromHapFreq = &(gpSample->StateEffectif[hapAlleleIndex[seqIDVec[dupHapIndex[dups]]]]);
		int *totToHapFreq = &(gpSample->StateEffectif[hapAlleleIndex[seqIDVec[origHapIndex[dups]]]]);
		if (augmentedHapSize != (*totFromHapFreq + *totToHapFreq)) {
			cerr << "\nError in void CNexus::mergeDupHapCounts() "
					"\naugmentedHapSize != (*totFromHapFreq + *totToHapFreq)"
					"\n Please contact the authors."
					"\nAborting Migraine...Press any key to exit.\n";
			if (cinGetOnError)
				cin.get();
			exit(-1);
		}
		else
			*totToHapFreq += *totFromHapFreq;

		if (DebugNexusSample)
			NexusMsgStr << endl << seqIDVec[dupHapIndex[dups]] << " -> " << seqIDVec[origHapIndex[dups]] << endl;

		//	updates for keeping count of the number of minor allele states for every gene
		if (outlierGeneFileBool && !promptHyperSegSites) {
			string fromSeq = seqIDVec[dupHapIndex[dups]], toSeq = seqIDVec[origHapIndex[dups]];
			if ((mapHap2Genes.find(fromSeq) == mapHap2Genes.end())
					|| (mapHap2Genes.find(toSeq) == mapHap2Genes.end())) {
				cerr << "\nError in void CNexus::mergeDupHapCounts() "
						"\n(mapHap2Genes.find(fromSeq) == mapHap2Genes.end()) || (mapHap2Genes.find(toSeq) == mapHap2Genes.end())"
						"\n Please contact the authors."
						"\nAborting Migraine...Press any key to exit.\n";
				if (cinGetOnError)
					cin.get();
				exit(-1);
			}
			else {
				for (vector<string>::iterator it = mapHap2Genes[fromSeq].begin(); it != mapHap2Genes[fromSeq].end(); ++it)
					mapHap2Genes[toSeq].push_back(*it);
				mapHap2Genes.erase(fromSeq);
			}
		}

		//	erase the duplicate haplotype
		removeHap(gpSample,dupHapIndex[dups],seqIDVec[dupHapIndex[dups]],true);
	}

}

void CNexus::removeHap(CSample *gpSample, int elimSeqIndex, string elimSeqLabel, bool dupSeq)
{
/*
	eliminates a single haplotype
*/

	NexusMsgStr << "Eliminating " << elimSeqLabel << "..." << endl;

	//	erasing the allele count from all pops
	for (size_t pop = 0; pop < gpSample->Effectif.size(); pop++) {
		//	checks if the about-to-be-eliminated haplotype is the last allele of the pop
		if (gpSample->Effectif[pop][hapAlleleIndex[elimSeqLabel]] < gpSample->PopEffectif[pop])
			gpSample->Effectif[pop].erase(gpSample->Effectif[pop].begin() + hapAlleleIndex[elimSeqLabel]);
		else {
			cerr << "\nCould not eliminate the sequence " << elimSeqLabel
				 <<	"\nas it is the only haplotype/allele present in pop number " << pop + 1
				 << "\nPlease restart by making a different choice during the elimination of sites/sequences."
					"\nAborting Migraine...Press any key to exit.\n";
			if (cinGetOnError)
				cin.get();
			exit(-1);
		}
		if (!dupSeq)
			gpSample->PopEffectif[pop] = accumulate(gpSample->Effectif[pop].begin(),gpSample->Effectif[pop].end(),0);
	}

	//	erasing the total count of this allele
	gpSample->StateEffectif.erase(gpSample->StateEffectif.begin() + hapAlleleIndex[elimSeqLabel]);

	//	erasing the allele index and its corresponding link with the Nexus sequences
	gpSample->alleleIndices.erase(gpSample->alleleIndices.begin() + hapAlleleIndex[elimSeqLabel]);
	hapAlleleIndex.erase(elimSeqLabel);

	//	correcting the size of the sample
	if (!dupSeq)
		gpSample->Size = accumulate(gpSample->PopEffectif.begin(),gpSample->PopEffectif.end(),0);

	//	decrease the allele number
	--gpSample->Sam_AlleleNbr;

	//	definitely erasing the sequence from the haplotype matrix
	seqIDVec.erase(find(seqIDVec.begin(),seqIDVec.end(),elimSeqLabel));
	vector<string> delStr(1,seqVec[elimSeqIndex]);
	seqVec.erase(find_end(seqVec.begin(),seqVec.end(),delStr.begin(),delStr.end()));
	seqNum = seqVec.size();

	//	updating the sequence-allele links
	mapNexusToGenepop(gpSample,true);

	NexusMsgStr << "\nNumber of informative sites : " << currInfSites << endl
		 << "Number of remaining haplotypes : " << seqNum << endl;
}

void CNexus::checkForPerfectPhylogeny()
{
/*
	Checks for the tree called perfect phylogeny, for sequence data presented in a binary format, which
	complies with the ISM model assumptions.
	The code below is borrowed from the program "Genetree" developed by R.C.Griffiths
	(http://www.stats.ox.ac.uk/~griff/software.html). See also "Efficient Algorithms for Inferring
	Evolutionary Trees" by Dan Gusfield (http://csiflabs.cs.ucdavis.edu/~gusfield/nperfect.pdf).
*/

	int siteCompArray[2][2];	//	a 4x4 array (counter) which keeps track of the binary configurations
								//	between a pair of sites and across all the haplotypes

	recodeISMBool = false;
	pass4GTBool = true;
	fourGameteTestCount.clear();
	fourGameteTestCount.resize(seqBinMat[0].size(),0);
	fourGameteTestPerHap.clear();
	fourGameteTestPerHap.resize(seqNum,0);

	//	Loops over all pairs of sites only once
	for (size_t siteIndex = 1; siteIndex < seqBinMat[0].size(); siteIndex++) {
		for (size_t subsiteIndex = 0; subsiteIndex < siteIndex; subsiteIndex++) {

			//	Loops over all four binary configurations: 00, 01, 10 and 11 to set siteCompArray[2][2] to zero
			for (int bin1 = 0; bin1 < 2; bin1++)
				for (int bin2 = 0; bin2 < 2; bin2++)
					siteCompArray[bin1][bin2] = 0;

			//	Loops over all the haplotypes
			for (size_t seqIndex = 0; seqIndex < seqNum; seqIndex++) {
				//	Once again looping over all four binary configurations: 00, 01, 10 and 11
				for (int bin1 = 0; bin1 < 2; bin1++)
					for (int bin2 = 0; bin2 < 2; bin2++)
						//	siteCompArray increments the configuration counters for a given pair of sites
						siteCompArray[bin1][bin2] += ((seqBinMat[seqIndex][siteIndex] == bin1)
															&& (seqBinMat[seqIndex][subsiteIndex] == bin2));
			}

			fourGameteTestMat[siteIndex][subsiteIndex] = fourGameteTestMat[subsiteIndex][siteIndex] = "-";

			//	IF for a given pair of sites only 01, 10 and 11 exist THEN ancestral and derived states need to be swapped
			if ((siteCompArray[0][1] != 0) && (siteCompArray[1][0] != 0) && (siteCompArray[1][1] != 0)) {
				fourGameteTestMat[siteIndex][subsiteIndex] = fourGameteTestMat[subsiteIndex][siteIndex] = "S";
				recodeISMBool = true;
			}

			//	IF for a given pair of sites all 4 configurations exist THEN no perfect phylogeny possible with this data
			if ((siteCompArray[0][1] != 0) && (siteCompArray[1][0] != 0)
					&& (siteCompArray[1][1] != 0) && (siteCompArray[0][0] != 0)) {
				fourGameteTestMat[siteIndex][subsiteIndex] = fourGameteTestMat[subsiteIndex][siteIndex] = "X";
				++fourGameteTestCount[siteIndex];
				++fourGameteTestCount[subsiteIndex];
				pass4GTBool = false;

				//	Loops over all the haplotypes
				for (size_t seqIndex = 0; seqIndex < seqNum; seqIndex++)
					//	checking for the binary configuration : 11
					if (seqBinMat[seqIndex][siteIndex] && seqBinMat[seqIndex][subsiteIndex])
						++fourGameteTestPerHap[seqIndex];
			}
		}
	}
}

void CNexus::printHaps(bool final)
{
	if (DebugNexusSample || final) {
		NexusMsgStr << "\nThe haplotype matrix :\n";
		for (size_t j = 0; j < seqSize; j++)
			if (siteState[ (int) j] == "CPP")//CM->RL 23.01.14
				NexusMsgStr << MRCASeq[j];
		NexusMsgStr << "\tAnc\n";

		for (size_t i = 0; i < seqNum; i++) {
			for (size_t j = 0; j < seqSize; j++)
				if (siteState[ (int) j] == "CPP")//CM->RL 23.01.14
					NexusMsgStr << seqVec[i][j];
			NexusMsgStr << "\t" << seqIDVec[i] << "\n";
		}

		NexusMsgStr << "\nThe binary version of the haplotype matrix :" << endl;
		for (size_t i = 0; i < seqNum; i++) {
			for (size_t j = 0; j < seqBinMat[0].size(); j++)
				NexusMsgStr << seqBinMat[i][j];
			NexusMsgStr << "\t" << seqIDVec[i] << "\n";
		}
	}

	if (!pass4GTBool || recodeISMBool) {
        cout << "\nCould not find a perfect phylogeny for this data set..." << endl;
        cout << "Check 'output_seqloc_job*.txt' for details." << endl;
		NexusMsgStr << "\nCould not find a perfect phylogeny for this data..."
				"\nThe crosses shown below highlight the pairwise incompatibilities that exist "
				"\n(and which need to be resolved) between any two nucleotide positions."
				"\nThe pairwise nucleotide four gamete test :" << endl;

		NexusMsgStr << "  ";
		for (size_t i = 0; i < fourGameteTestMat.size(); i++)
			NexusMsgStr << (i + 1) % 10 << " ";
		NexusMsgStr << "\t\"X\" counts" << endl;

		for (size_t i = 0; i < fourGameteTestMat.size(); i++) {
			NexusMsgStr << (i + 1) % 10;
			for (size_t j = 0; j < fourGameteTestMat.size(); j++)
				NexusMsgStr << " " << fourGameteTestMat[i][j];
			NexusMsgStr << " \t    " << fourGameteTestCount[i] << endl;
		}
		NexusMsgStr << "\nTotal number of \"X\" counts/incompatibilities : " << accumulate(fourGameteTestCount.begin(),fourGameteTestCount.end(),0) << endl;
	}
}

void CNexus::recodeISMSeq()
{
/*
	This function is called when a perfect phylogeny is possible from the haplotype data in the NEXUS file
	but with some changes. The latter consists of swapping the ancestral and derived states of sites when
	the derived allele is found more abundantly than the ancestral one.
	The code below is borrowed from the program "Genetree" developed by R.C.Griffiths
	(http://www.stats.ox.ac.uk/~griff/software.html).
*/

	int siteCount[2];	//	the counter for the ancestral/derived states at a segregating site

	//	Loops over all valid sites
	for (size_t siteIndex = 0; siteIndex < seqSize; siteIndex++) {
		if (siteState[ (int) siteIndex] == "CPP") {//CM->RL 23.01.14
			siteCount[0] = 0;
			siteCount[1] = 0;

			//	increments the counter w.r.t. the state encountered
			for (size_t seqIndex = 0; seqIndex < seqNum; seqIndex++)
				++siteCount[seqBinMat[seqIndex][siteIndex]];

			//	IF the derived allele is more abundant then...SWAP the state!
			if (siteCount[1] > siteCount[0])
				for (size_t seqIndex = 0; seqIndex < seqNum; seqIndex++)
					seqBinMat[seqIndex][siteIndex] = 1 - seqBinMat[seqIndex][siteIndex];
		}
	}
}

void CStoreNexus::copyCSample(CSample *gpSample)
{
	alleleIndices = gpSample->alleleIndices;
	StateEffectif = gpSample->StateEffectif;
	Effectif = gpSample->Effectif;
	PopEffectif = gpSample->PopEffectif;
	Size = gpSample->Size;
	Sam_AlleleNbr = gpSample->Sam_AlleleNbr;
    mutPathMatrix = gpSample->mutPathMatrix;
	siteMutCountVec = gpSample->siteMutCountVec;
	DmVec = gpSample->DmVec;
}

void CStoreNexus::getCSample(CSample *gpSample)
{
	gpSample->alleleIndices = alleleIndices;
	gpSample->StateEffectif = StateEffectif;
	gpSample->Effectif = Effectif;
	gpSample->PopEffectif = PopEffectif;
	gpSample->Size = Size;
	gpSample->Sam_AlleleNbr = Sam_AlleleNbr;
	gpSample->mutPathMatrix = mutPathMatrix;
	gpSample->siteMutCountVec = siteMutCountVec;
	gpSample->DmVec = DmVec;
}

