/***************************************************************************
© F. Rousset, R. Leblois, C.R. Beeravolu 2012-

francois.rousset@univ-montp2.fr
raphael.leblois@mnhn.fr

This file is part of Migraine. This software is a computer program
whose purpose is to perform statistical analysis.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***************************************************************************/

#ifndef NEXUS_H_
#define NEXUS_H_

#include "dgCSample.h"

using namespace std;

//	A class for storing quantities which need to re-attributed to a CSample for every call to calclogl
class CStoreNexus
{
private:
	//	copies of CSample members which which need to be initialised only once during the NEXUS procedure
	vector<int> alleleIndices;
	vector<int> StateEffectif;
	vector<vector<int> > Effectif;
	vector<int> PopEffectif;
	int Size;
	int Sam_AlleleNbr;
	vector<vector<int> > mutPathMatrix;
	vector<int> siteMutCountVec;
	vector<int> DmVec;

	void copyCSample(CSample *gpSample);

public:
	void getCSample(CSample *gpSample);
	friend class CNexus;
};

//	A class for reading and extracting data of and from files in the NEXUS format respectively
class CNexus
{
private:
	string outFilename;						//	the filename for outputing messages related to the NEXUS file processing
	stringstream NexusMsgStr;				//	the output message string stream
	string lineBuf; 			 			// 	reads the whole of the NEXUS file into a "single line"
	bool nexusFileOK; 						// 	indicates whether the file was opened correctly and "cleaned" if necessary
	bool recodeISMBool;						// 	is true when the ancestral and derived states at a nucleotide need to be shuffled
	bool pass4GTBool;						// 	result of the 4 gamete (pairwise nucleotide) test across all haplotypes
	size_t seqSize;							// 	the "nucleotide" dimension of the haplotype matrix (includes the incompatible sites)
	size_t seqNum;							// 	the "sequences" dimension of the haplotype matrix (minus the duplicates)
	size_t currInfSites;					//	current number of informative sites
	size_t origInfSites;					//	original number of informative sites (includes hyper-segregating sites)
	int hyperSegCount;						// 	a counter of the number of nucleotides having more than two alleles
	int samSize;							//	the original sample size in terms of the number of genes

	vector<char> MRCASeq; 					//	the MRCA or reference sequence
	vector<string> seqIDVec;				// 	the haplotype labels matrix
	vector<string> seqVec; 					// 	the haplotype matrix in ATGC format
	vector<string> modSeqVec;				// 	a temporary MODIFIED haplotype matrix (contains only the ISM compatible sites)
	vector<int> modHapMatSiteIndex;			// 	maintains the link between the site indices of the modified and unmodified haplotype matrices
	vector<vector<int> > mutPathMatrix;		// 	contains the tree version (mutation paths to the MRCA) of the haplotype matrix
	vector<vector<int> > seqBinMat;			// 	the modified haplotype matrix in a binary format
	vector<vector<string> > fourGameteTestMat; 	//	the site by site matrix containing the results of the four gamete test (4GT)
	vector<int> fourGameteTestCount;		// 	the site-wise count of 4GT incompatibilities
	vector<int> fourGameteTestPerHap;		// 	the haplotype-wise count of 4GT incompatibilities
	map<string, int> hapAlleleIndex;		//	maintains the link between every Nexus taxon label to its corresponding allele index (i.e. CSample::alleleIndices[index])
	map<string, vector<int> > outlierGeneCount;	//	counts the number of minor allele states and 4GT incompatibilities for each gene
	map<string, vector<string> > mapHap2Genes; // maps every haplotype (or unique sequence) ID to its corresponding gene/sequence IDs
	vector<int> siteMutCountVec;			// 	a counter of the number of allelic states at each nucleotide site
	vector<int> DmVec;						//  the number of SEQUENCES containing a given mutation at a nucleotide position
	map<int, map<char, int> > siteStatesCount;	//	A,T,G,C counts at each nucleotide site/position
	map<int, string> siteState;				//	records the state of each nucleotide as follows:
											//	UN - Uninitialised;
											//	NS/HS - non-segregating/hyper segregating or
											//	CPP/NPP - check for perfect phylogeny/NO perfect phylogeny

	void removeNEXUSComments();
	void removeDuplicates();
	void mapNexusToGenepop(CSample *gpSample, bool update);
	void countNucStates();
	void setMRCA();
	void processHyperSegSites(CSample *gpSample);
	void mergeDupHapCounts(CSample *gpSample, bool outlierBool);
	void removeHap(CSample *gpSample, int elimSeqIndex, string elimSeqLabel, bool dupSeq);
	void checkForPerfectPhylogeny();
	void printHaps(bool final);
	void recodeISMSeq();

public:
	int locNum; 						// 	the locus number currently being considered by Migraine (numbering starts with 1 and not zero!)
	size_t hapCount; 					// 	the original number of haplotypes (unique sequences) present in the NEXUS file (ideally every sequence should be unique)
	string fileName;	 				// 	the NEXUS filename containing the haplotypes at the "locNum" locus
	CStoreNexus CopyCatNexus;

	CNexus();
	void cleanNEXUSFile(const int &locusNum);
	void readNEXUSFile(CSample *gpSample);
	void checkIfDataISMCompatible(CSample *gpSample);
    void FillHaplotypeInfoFromSNP(CSample *gpSample);

};

#endif /* NEXUS_H_ */


